
#include "Mt2/Elliptic_Mt2_332_Calculator.h"
#include "Mt2/Mt2Units.h"

namespace Mt2 {

  Elliptic_Mt2_332_Calculator::Elliptic_Mt2_332_Calculator() : Mt2_332_Calculator("Elliptic_Mt2_332") {
    solved=false;
    SCANSTEP=0.5;
    MIN_MASS=1E-6;
    PRECISION=1E-3;
  }

  double Elliptic_Mt2_332_Calculator::mt2_332_Sq(const LorentzTransverseVector& visibleA, // 3 d.o.f. 
						 const LorentzTransverseVector& visibleB, // 3 d.o.f.
						 const TwoVector& ptmiss,                 // 2 d.o.f.
						 const double mInvisible){

    double mt2=Elliptic_Mt2_332_Calculator::mt2_332(visibleA, visibleB, ptmiss, mInvisible);
    return mt2*mt2;
  }
  

  double Elliptic_Mt2_332_Calculator::mt2_332(const LorentzTransverseVector& visA, 
					      const LorentzTransverseVector& visB,
					      const TwoVector& ptmiss, 
					      const double mEachInvisible) {
    // set momenta
    needscan=0;
    solved = false;
    mt2_b=0.;

    ma = visA.mass();
    pax = visA.px();
    pay = visA.py();
    masq = ma*ma;
    Easq = masq+pax*pax+pay*pay;
    Ea   = sqrt(Easq);
    
    mb = visB.mass();
    pbx  = visB.px(); 
    pby  = visB.py();
    mbsq = mb*mb;
    Ebsq = mbsq+pbx*pbx+pby*pby;
    Eb   = sqrt(Ebsq);
     
    pmissx   = ptmiss.px(); 
    pmissy   = ptmiss.py();
    pmissxsq = pmissx*pmissx;
    pmissysq = pmissy*pmissy;

// set ma>= mb
   if(masq < mbsq)
   {
      double temp;
      temp = pax;  pax  = pbx;  pbx  = temp;
      temp = pay;  pay  = pby;  pby  = temp;
      temp = Ea;   Ea   = Eb;   Eb   = temp;
      temp = Easq; Easq = Ebsq; Ebsq = temp;
      temp = masq; masq = mbsq; mbsq = temp;
      temp = ma;   ma   = mb;   mb   = temp;   
   }

    mn = mEachInvisible;
    mnsq = mn*mn;

    mt2_bisect();
    return mt2_b;
  }


  //special case, the visible particle is massless
void Elliptic_Mt2_332_Calculator::mt2_massless()
{
  using namespace std;
   
//rotate so that pay = 0 
   double theta,s,c;
   theta = atan(pay/pax);
   s = sin(theta);
   c = cos(theta);

   double pxtemp,pytemp;
   Easq   = pax*pax+pay*pay;
   Ebsq   = pbx*pbx+pby*pby;
   Ea     = sqrt(Easq);
   Eb     = sqrt(Ebsq);
  
   pxtemp = pax*c+pay*s;
   pax    = pxtemp;
   pay    = 0;
   pxtemp = pbx*c+pby*s;
   pytemp = -s*pbx+c*pby;
   pbx    = pxtemp;
   pby    = pytemp;
   pxtemp = pmissx*c+pmissy*s;
   pytemp = -s*pmissx+c*pmissy;
   pmissx = pxtemp;
   pmissy = pytemp;

   a2  = 1-pbx*pbx/(Ebsq);
   b2  = -pbx*pby/(Ebsq);
   c2  = 1-pby*pby/(Ebsq);
 
   d21 = (Easq*pbx)/Ebsq;
   d20 = - pmissx +  (pbx*(pbx*pmissx + pby*pmissy))/Ebsq;
   e21 = (Easq*pby)/Ebsq;
   e20 = - pmissy +  (pby*(pbx*pmissx + pby*pmissy))/Ebsq;
   f22 = -(Easq*Easq/Ebsq);
   f21 = -2*Easq*(pbx*pmissx + pby*pmissy)/Ebsq;
   f20 = mnsq + pmissxsq + pmissysq - pow((pbx*pmissx + pby*pmissy)/Eb,2);

   double Deltasq0    = 0;
   double Deltasq_low, Deltasq_high;
   int    nsols_high, nsols_low;

   Deltasq_low = Deltasq0 + 1;
   nsols_low = nsols_massless(Deltasq_low);
   
   if(nsols_low > 1) 
   { 
      mt2_b = (double) sqrt(Deltasq0+mnsq);
      return;
   }

/*   
   if( nsols_massless(Deltasq_high) > 0 )
   {
      mt2_b = (double) sqrt(mnsq+Deltasq0);
      return;
   }*/

//look for when both parablos contain origin  
   double Deltasq_high1, Deltasq_high2;
   Deltasq_high1 = 2*Eb*sqrt(pmissx*pmissx+pmissy*pmissy+mnsq)-2*pbx*pmissx-2*pby*pmissy;
   Deltasq_high2 = 2*Ea*mn;
 
   if(Deltasq_high1 < Deltasq_high2) Deltasq_high = Deltasq_high2;
     else Deltasq_high = Deltasq_high1;

   nsols_high=nsols_massless(Deltasq_high);
  
   int foundhigh;
   if (nsols_high == nsols_low)
   {
      needscan++;
      
      foundhigh=0;
      
      double minmass, maxmass;
      minmass  = mn ;
      maxmass  = sqrt(mnsq+Deltasq_high);
      for(double mass = minmass + SCANSTEP; mass < maxmass; mass += SCANSTEP)
      {
	 Deltasq_high = mass*mass - mnsq;
      
         nsols_high = nsols_massless(Deltasq_high);
         if(nsols_high>0)
         {
            foundhigh=1;
            Deltasq_low = (mass-SCANSTEP)*(mass-SCANSTEP) - mnsq;
            break;
         }
      }
      if(foundhigh==0) 
      {
       
	  cout<<"Deltasq_high not found"<<endl;
        
       
         mt2_b = (double)sqrt(Deltasq_low+mnsq);
         return;
      }
   }

   if(nsols_high == nsols_low)
   { 
      cout << "error: nsols_low=nsols_high=" << nsols_high << endl;
      cout << "Deltasq_high=" << Deltasq_high << endl;
      cout << "Deltasq_low= "<< Deltasq_low << endl;
    
      mt2_b = sqrt(mnsq+Deltasq_low);
      return;
   }
   double minmass, maxmass;
   minmass = sqrt(Deltasq_low+mnsq);
   maxmass = sqrt(Deltasq_high+mnsq);
   while(maxmass - minmass > PRECISION)
   {
      double Delta_mid, midmass, nsols_mid;
      midmass   = (minmass+maxmass)/2.;
      Delta_mid = midmass * midmass - mnsq;
      nsols_mid = nsols_massless(Delta_mid);
      if(nsols_mid != nsols_low) maxmass = midmass;
      if(nsols_mid == nsols_low) minmass = midmass;
   }
   mt2_b = minmass;
   return;
}

int Elliptic_Mt2_332_Calculator::nsols_massless(double Dsq)
{
  double delta;
  delta = Dsq/(2*Easq);
  d1    = d11*delta;
  e1    = e11*delta;
  f1    = f12*delta*delta+f10;
  d2    = d21*delta+d20;
  e2    = e21*delta+e20;
  f2    = f22*delta*delta+f21*delta+f20;
  
    double a,b;
  if (pax>0) a = Ea/Dsq;
  else       a = -Ea/Dsq;
  if (pax>0) b = -Dsq/(4*Ea)+mnsq*Ea/Dsq;
  else       b = Dsq/(4*Ea)-mnsq*Ea/Dsq;
  
  double A4,A3,A2,A1,A0;

  A4 = a*a*a2;
  A3 = 2*a*b2/Ea;
  A2 = (2*a*a2*b+c2+2*a*d2)/(Easq);
  A1 = (2*b*b2+2*e2)/(Ea*Easq);
  A0 = (a2*b*b+2*b*d2+f2)/(Easq*Easq);

  long  double t1,t2,t3,t4,t5;
  t1 = A4;
  t2 = A4;
  t3 = -(A2/2-3*A3*A3/(16*A4));
  t4 = -(A4*(36*pow(A1,2)*pow(A4,2) + 2*pow(A2,2)*(-pow(A3,2) + 4*A2*A4) + 
      A1*(6*pow(A3,3) - 28*A2*A3*A4) + 
      A0*(12*pow(A3,2)*A4 - 32*A2*pow(A4,2))));
  
  t5 = -(pow(3*pow(A3,2) - 8*A2*A4,2)*
     (27*pow(A1,4)*pow(A4,2) +
       2*pow(A1,3)*A3*(2*pow(A3,2) - 9*A2*A4) +
       2*A0*A1*A3*(-9*A2*pow(A3,2) + 40*pow(A2,2)*A4 +
          96*A0*pow(A4,2)) +
       pow(A1,2)*(-(pow(A2,2)*pow(A3,2)) + 4*pow(A2,3)*A4 +
          6*A0*pow(A3,2)*A4 - 144*A0*A2*pow(A4,2)) +
       A0*(4*pow(A2,3)*pow(A3,2) + 27*A0*pow(A3,4) -
          16*A2*(pow(A2,3) + 9*A0*pow(A3,2))*A4 +
	   128*A0*pow(A2,2)*pow(A4,2) - 256*pow(A0,2)*pow(A4,3))));
   int nsol;
   nsol = signchange_n(t1,t2,t3,t4,t5)-signchange_p(t1,t2,t3,t4,t5);
   if( nsol < 0 ) nsol=0;

   return nsol;
  
}

void Elliptic_Mt2_332_Calculator::mt2_bisect()
{
  using namespace std;
   
//if masses are very small, use code for massless case.  
   if(masq < MIN_MASS && mbsq < MIN_MASS) 
   { 
      mt2_massless();
      return;
   }
 

   double Deltasq0;     
   Deltasq0=ma*(ma+2*mn); //The minimum mass square to have two ellipses 
 
// find the coefficients for the two quadratic equations when Deltasq=Deltasq0.
  
   a1 = 1-pax*pax/(Easq);
   b1 = -pax*pay/(Easq);
   c1 = 1-pay*pay/(Easq);
   d1 = -pax*(Deltasq0-masq)/(2*Easq);
   e1 = -pay*(Deltasq0-masq)/(2*Easq);
   a2 = 1-pbx*pbx/(Ebsq);
   b2 = -pbx*pby/(Ebsq);
   c2 = 1-pby*pby/(Ebsq);
   d2 = -pmissx+pbx*(Deltasq0-mbsq)/(2*Ebsq)+pbx*(pbx*pmissx+pby*pmissy)/(Ebsq);
   e2 = -pmissy+pby*(Deltasq0-mbsq)/(2*Ebsq)+pby*(pbx*pmissx+pby*pmissy)/(Ebsq);
   f2 = pmissx*pmissx+pmissy*pmissy-((Deltasq0-mbsq)/(2*Eb)+
        (pbx*pmissx+pby*pmissy)/Eb)*((Deltasq0-mbsq)/(2*Eb)+
        (pbx*pmissx+pby*pmissy)/Eb)+mnsq;
   
// find the center of the smaller ellipse 
   double x0,y0;
   x0 = (c1*d1-b1*e1)/(b1*b1-a1*c1);
   y0 = (a1*e1-b1*d1)/(b1*b1-a1*c1);

   
// Does the larger ellipse contain the smaller one? 
   double dis=a2*x0*x0+2*b2*x0*y0+c2*y0*y0+2*d2*x0+2*e2*y0+f2;

   if(dis<=0.01)
   { 
      mt2_b  = (double) sqrt(mnsq+Deltasq0);
      return;
   }
   

/* find the coefficients for the two quadratic equations           */
/* coefficients for quadratic terms do not change                  */
/* coefficients for linear and constant terms are polynomials of   */
/*       delta=(Deltasq-m7sq)/(2 E7sq)                             */  
   d11 = -pax;
   e11 = -pay;
   f10 = mnsq;
   f12 = -Easq;
   d21 = (Easq*pbx)/Ebsq;
   d20 = ((masq - mbsq)*pbx)/(2.*Ebsq) - pmissx +
         (pbx*(pbx*pmissx + pby*pmissy))/Ebsq;
   e21 = (Easq*pby)/Ebsq;
   e20 = ((masq - mbsq)*pby)/(2.*Ebsq) - pmissy +
         (pby*(pbx*pmissx + pby*pmissy))/Ebsq;
   f22 = -Easq*Easq/Ebsq;
   f21 = (-2*Easq*((masq - mbsq)/(2.*Eb) + (pbx*pmissx + pby*pmissy)/Eb))/Eb;
   f20 = mnsq + pmissx*pmissx + pmissy*pmissy - 
         pow((masq - mbsq)/(2.*Eb) + (pbx*pmissx + pby*pmissy)/Eb,2);

//Estimate upper bound of mT2
//when Deltasq > Deltasq_high1, the larger encloses the center of the smaller 
   double p2x0,p2y0;
   double Deltasq_high1;
   p2x0 = pmissx-x0;
   p2y0 = pmissy-y0;
   Deltasq_high1 = 2*Eb*sqrt(p2x0*p2x0+p2y0*p2y0+mnsq)-2*pbx*p2x0-2*pby*p2y0+mbsq;
   
//Another estimate, if both ellipses enclose the origin, Deltasq > mT2

   double Deltasq_high2, Deltasq_high21, Deltasq_high22;
   Deltasq_high21 = 2*Eb*sqrt(pmissx*pmissx+pmissy*pmissy+mnsq)-2*pbx*pmissx-2*pby*pmissy+mbsq;
   Deltasq_high22 = 2*Ea*mn+masq;
  
   if ( Deltasq_high21 < Deltasq_high22 ) Deltasq_high2 = Deltasq_high22;
   else Deltasq_high2 = Deltasq_high21;

//pick the smaller upper bound   
   double Deltasq_high;
   if(Deltasq_high1 < Deltasq_high2) Deltasq_high = Deltasq_high1;
   else Deltasq_high = Deltasq_high2;
   
  
   double Deltasq_low; //lower bound
   Deltasq_low = Deltasq0;

//number of solutions at Deltasq_low should not be larger than zero
   if( nsols(Deltasq_low) > 0 )
   {
     //cout << "nsolutions(Deltasq_low) > 0"<<endl;
     mt2_b = (double) sqrt(mnsq+Deltasq0);
     return;
   }
  
   int nsols_high, nsols_low;

   nsols_low  = nsols(Deltasq_low);
   int foundhigh;
  

//if nsols_high=nsols_low, we missed the region where the two ellipse overlap 
//if nsols_high=4, also need a scan because we may find the wrong tangent point.

   nsols_high = nsols(Deltasq_high);
  
   if(nsols_high == nsols_low || nsols_high == 4)
   {
      foundhigh = scan_high(Deltasq_high);
      if(foundhigh == 0) 
      {
	 cout<<"Deltasq_high not found"<<endl;
         mt2_b = sqrt( Deltasq_low + mnsq );
         return;
      }
      
   }

   while(sqrt(Deltasq_high+mnsq) - sqrt(Deltasq_low+mnsq) > PRECISION)
   {
      double Deltasq_mid,nsols_mid;
      //bisect
      Deltasq_mid = (Deltasq_high+Deltasq_low)/2.;
      nsols_mid = nsols(Deltasq_mid);
      // if nsols_mid = 4, rescan for Deltasq_high
      if ( nsols_mid == 4 ) 
      {
         Deltasq_high = Deltasq_mid;
         scan_high(Deltasq_high);
         continue;
      } 
         
      
      if(nsols_mid != nsols_low) Deltasq_high = Deltasq_mid;
      if(nsols_mid == nsols_low) Deltasq_low  = Deltasq_mid;
   }
   mt2_b = (double) sqrt( mnsq + Deltasq_high);
   return;
}


int Elliptic_Mt2_332_Calculator::scan_high(double & Deltasq_high)
{
   int foundhigh = 0 ;
   int nsols_high;

   //double Deltasq_high1 = Deltasq_high;
   double Deltasq_low;
   double tempmass, maxmass;
   tempmass = mn + ma;
   maxmass  = sqrt(mnsq+Deltasq_high);
  
   for(double mass = tempmass + SCANSTEP; mass < maxmass; mass += SCANSTEP)
   {
      Deltasq_high = mass*mass - mnsq;
      nsols_high   = nsols(Deltasq_high);
      
      if( nsols_high > 0)
      {
	 Deltasq_low = (mass-SCANSTEP)*(mass-SCANSTEP) - mnsq;
         foundhigh   = 1;
         break;
      }
    }
    return foundhigh;
}
int Elliptic_Mt2_332_Calculator::nsols(  double Dsq)
{
   double delta = (Dsq-masq)/(2*Easq);
  
//calculate coefficients for the two quadratic equations
   d1 = d11*delta;
   e1 = e11*delta;
   f1 = f12*delta*delta+f10;
   d2 = d21*delta+d20;
   e2 = e21*delta+e20;
   f2 = f22*delta*delta+f21*delta+f20;

//obtain the coefficients for the 4th order equation 
//devided by Ea^n to make the variable dimensionless
   long double A4, A3, A2, A1, A0;


   A4 = 
   -4*a2*b1*b2*c1 + 4*a1*b2*b2*c1 +a2*a2*c1*c1 + 
   4*a2*b1*b1*c2 - 4*a1*b1*b2*c2 - 2*a1*a2*c1*c2 + 
   a1*a1*c2*c2;  

   A3 =
     (-4*a2*b2*c1*d1 + 8*a2*b1*c2*d1 - 4*a1*b2*c2*d1 - 4*a2*b1*c1*d2 + 
   8*a1*b2*c1*d2 - 4*a1*b1*c2*d2 - 8*a2*b1*b2*e1 + 8*a1*b2*b2*e1 + 
   4*a2*a2*c1*e1 - 4*a1*a2*c2*e1 + 8*a2*b1*b1*e2 - 8*a1*b1*b2*e2 - 
     4*a1*a2*c1*e2 + 4*a1*a1*c2*e2)/Ea;

   A2 =
     (4*a2*c2*pow(d1,2) - 4*a2*c1*d1*d2 - 4*a1*c2*d1*d2 + 4*a1*c1*pow(d2,2) - 
   8*a2*b2*d1*e1 - 8*a2*b1*d2*e1 + 16*a1*b2*d2*e1 + 
   4*pow(a2,2)*pow(e1,2) + 16*a2*b1*d1*e2 - 8*a1*b2*d1*e2 - 
   8*a1*b1*d2*e2 - 8*a1*a2*e1*e2 + 4*pow(a1,2)*pow(e2,2) - 4*a2*b1*b2*f1 + 
   4*a1*pow(b2,2)*f1 + 2*pow(a2,2)*c1*f1 - 2*a1*a2*c2*f1 + 
     4*a2*pow(b1,2)*f2 - 4*a1*b1*b2*f2 - 2*a1*a2*c1*f2 + 2*pow(a1,2)*c2*f2)/(Easq);

   A1 =
     (-8*a2*d1*d2*e1 + 8*a1*pow(d2,2)*e1 + 8*a2*pow(d1,2)*e2 - 8*a1*d1*d2*e2 - 
   4*a2*b2*d1*f1 - 4*a2*b1*d2*f1 + 8*a1*b2*d2*f1 + 4*pow(a2,2)*e1*f1 - 
   4*a1*a2*e2*f1 + 8*a2*b1*d1*f2 - 4*a1*b2*d1*f2 - 4*a1*b1*d2*f2 - 
     4*a1*a2*e1*f2 + 4*pow(a1,2)*e2*f2)/(Easq*Ea);
  
   A0 =
     (-4*a2*d1*d2*f1 + 4*a1*pow(d2,2)*f1 + pow(a2,2)*pow(f1,2) + 
   4*a2*pow(d1,2)*f2 - 4*a1*d1*d2*f2 - 2*a1*a2*f1*f2 + 
     pow(a1,2)*pow(f2,2))/(Easq*Easq);

   long  double t1,t2,t3,t4,t5;
//find the coefficients for the leading term in the Sturm sequence  
   t1 = A4;
   t2 = A4;
   t3 = -(A2/2-3*A3*A3/(16*A4));
   t4 =
       -(A4*(36*pow(A1,2)*pow(A4,2) + 2*pow(A2,2)*(-pow(A3,2) + 4*A2*A4) + 
       A1*(6*pow(A3,3) - 28*A2*A3*A4) + 
       A0*(12*pow(A3,2)*A4 - 32*A2*pow(A4,2))));
  
   t5 = -(pow(3*pow(A3,2) - 8*A2*A4,2)*
       (27*pow(A1,4)*pow(A4,2) +
       2*pow(A1,3)*A3*(2*pow(A3,2) - 9*A2*A4) +
       2*A0*A1*A3*(-9*A2*pow(A3,2) + 40*pow(A2,2)*A4 +
          96*A0*pow(A4,2)) +
       pow(A1,2)*(-(pow(A2,2)*pow(A3,2)) + 4*pow(A2,3)*A4 +
          6*A0*pow(A3,2)*A4 - 144*A0*A2*pow(A4,2)) +
       A0*(4*pow(A2,3)*pow(A3,2) + 27*A0*pow(A3,4) -
          16*A2*(pow(A2,3) + 9*A0*pow(A3,2))*A4 +
	   128*A0*pow(A2,2)*pow(A4,2) - 256*pow(A0,2)*pow(A4,3))));

//The number of solutions depends on diffence of number of sign changes for x->Inf and x->-Inf
   int nsol;
   nsol = signchange_n(t1,t2,t3,t4,t5) - signchange_p(t1,t2,t3,t4,t5);

//Cannot have negative number of solutions, must be roundoff effect
   if (nsol < 0) nsol = 0;

   return nsol;
  
}  

inline int Elliptic_Mt2_332_Calculator::signchange_n( long double t1, long double t2, long double t3, long double t4, long double t5)
{
   int nsc;
   nsc=0;
   if(t1*t2>0) nsc++;
   if(t2*t3>0) nsc++;
   if(t3*t4>0) nsc++;
   if(t4*t5>0) nsc++;
   return nsc;
}
inline int Elliptic_Mt2_332_Calculator::signchange_p( long double t1, long double t2, long double t3, long double t4, long double t5)
{
   int nsc;
   nsc=0;
   if(t1*t2<0) nsc++;
   if(t2*t3<0) nsc++;
   if(t3*t4<0) nsc++;
   if(t4*t5<0) nsc++;
   return nsc;
}



} // end of Mt2 Namespace
