#include "Mt2/Mt2MinimiserGsl.h"
#include "Mt2/Mt2Calculator.h"
#include <gsl/gsl_min.h>
#include <gsl/gsl_errno.h>

namespace Mt2 {

Mt2MinimiserGsl::Mt2MinimiserGsl() : 
  m_tolerance (1E-8),
  m_max_fun (100)
{
}

Mt2MinimiserGsl::~Mt2MinimiserGsl()
{
}

double Mt2MinimiserGsl::minimise(const Mt2Calculator& calc) {
  Params params;
  params.calc = &calc;
  int status;
  int iter = 0, max_iter = m_max_fun;
  const gsl_min_fminimizer_type *T;
  gsl_min_fminimizer *s;
  
  const double par_min = 0.;
  const double par_max = calc.rootSMax-calc.rootSMin;
  if (par_max<par_min) {
    throw EventIsIncompatibleWithRootS_Exception();
  }

  const double fpar_min  = calc.trialMt2(par_min);
  const double fpar_max = calc.trialMt2(par_max);
  if (!finite(fpar_max)) {
    throw EventIsAlmostIncompatibleWithRootS_Exception();
  }

  // gsl demands that  f(par_min) > f(guess) < f(par_max).
  // We must now make sure that that is true!
  // In order to satisfy the above condition, we assume that f(x)
  // looks like a "Nike" swoosh between par_min and par_max.
  // If this is not the case, the algorithm will fail.

  int count=0;
  double guess_frac = 0.125;
  double guess;
  while(true) {
    guess = par_min*(1.-guess_frac) + par_max*(guess_frac);
    const double fguess = calc.trialMt2(guess);
    if (fpar_min > fguess && fguess < fpar_max) break;
    if (guess_frac==0 || count>100) {
      // failed to find valid value for "guess"
      throw ProgramWasNotCleverEnoughException("Unable to find a suitable start point for the GSL minimiser.  Calculation of MT2 failed.  Perhaps the \"Nike\" swoosh assumption is wrong.");
    }
    guess_frac*=0.125;
    ++count;
  }

  // By now we should have met the gsl demand that  f(par_min) > f(guess) < f(par_max).

  double m = guess; // m plays two roles.  (1) it is an initial x-value to be used when minimisation begins (satisfying the above conditions).  (2) in subsequent iterations it is modified by the gsl functions until it becomes the x-value of the minimum. It ought to be called x.

  double a = par_min;
  double b = par_max;
  gsl_function F;
  
  F.function = &Mt2MinimiserGsl::static_funct;
  F.params = (void*) &params;
  
  T = gsl_min_fminimizer_brent;
  s = gsl_min_fminimizer_alloc (T);
  printf("m a b %f %f %f\n",m,a,b);
  gsl_min_fminimizer_set (s, &F, m, a, b);

  if (m_debug) {
  printf ("using %s method\n",
	  gsl_min_fminimizer_name (s));
  
  printf ("%5s [%9s, %9s] %9s %9s\n",
	  "iter", "lower", "upper", "min",
	  "err(est)");
  
  printf ("%5d [%.7f, %.7f] %.7f %.7f\n",
	  iter, a, b,
	  m, b - a);
  }

  do
    {
      iter++;
      status = gsl_min_fminimizer_iterate (s);
      
      m = gsl_min_fminimizer_x_minimum (s);
      a = gsl_min_fminimizer_x_lower (s);
      b = gsl_min_fminimizer_x_upper (s);
      
      status 
	= gsl_min_test_interval (a, b, m_tolerance, 0.0);
      
      if (m_debug) {
	if (status == GSL_SUCCESS)
	printf ("Converged:\n");
      
	printf ("%5d [%.7f, %.7f] "
	"%.7f %.7f\n",
	iter, a, b,
	m, b - a);
      }
    }
  while (status == GSL_CONTINUE && iter < max_iter);
  
  gsl_min_fminimizer_free(s);

  return static_funct(m, F.params);
}

double Mt2MinimiserGsl::static_funct(double x, void* params) {
  Params * paramsProper = (Params*)params;
  const double ans = paramsProper->calc->trialMt2(x);
  return ans;
}

void Mt2MinimiserGsl::setTolerance(double t){
  m_tolerance = t;
}

void Mt2MinimiserGsl::setMaxEvaluations(int i){
  m_max_fun = i;
}

}
